
from generallibrary import Markdown, floor, Date, deco_cache, ObjInfo, flatten, exclusive
from generalfile import Path

import re


class _PackagerMarkdown:
    """ Contains methods to generate readme sections from arguments. """
    def get_badges_dict(self):
        """ Get badges as a dict.

            :param generalpackager.Packager self: """
        return {
            "UnitTests": f"[![workflow Actions Status](https://github.com/ManderaGeneral/{self.name}/workflows/workflow/badge.svg)](https://github.com/ManderaGeneral/{self.name}/actions)",
            "Commit": f"![GitHub last commit](https://img.shields.io/github/last-commit/ManderaGeneral/{self.name})",
            "Release": f"[![PyPI version shields.io](https://img.shields.io/pypi/v/{self.name}.svg)](https://pypi.org/project/{self.name}/)",
            "Python": f"[![PyPI pyversions](https://img.shields.io/pypi/pyversions/{self.name}.svg)](https://pypi.python.org/pypi/{self.name}/)",
            "Operating System": f"[![Generic badge](https://img.shields.io/badge/platforms-{'%20%7C%20'.join(self.os)}-blue.svg)](https://shields.io/)",
        }


    def _create_todo_dict(self, text, path, line):
        """ :param generalpackager.Packager self: """
        path = Path(path)
        return {
            "Package": Markdown.link(text=self.name, url=self.github.url),
            "Module": self.github_link_path_line(text=path.name(), path=path, line=1),
            "Message": self.github_link_path_line(text=text, path=path, line=line),
        }

    def _get_codeline_todos(self):
        """ :param generalpackager.Packager self: """
        todos = []
        for path in self.path.get_children(depth=-1, gen=True):
            if path.match(*self.git_exclude_lines, "shelved.patch", "readme.md"):
                continue
            try:
                text = path.text.read()
            except:
                continue

            relative_path = path.relative(self.path)

            for i, line in enumerate(text.splitlines()):
                result = re.findall("todo+: (.+)", line, re.I)
                if result:
                    text = re.sub('[" ]*$', "", result[0])
                    todos.append(self._create_todo_dict(text=text, path=relative_path, line=i + 1))
        return todos

    def _get_untested_todos(self):
        """ :param generalpackager.Packager self: """
        todos = []
        for name, objInfo in self.get_untested_objInfo_dict().items():  # type: ObjInfo
            text = f"[UnitTest] for {repr(objInfo)}"
            path = objInfo.file(relative=True)
            line = objInfo.get_definition_line()
            todos.append(self._create_todo_dict(text=text, path=path, line=line))
        return todos

    @deco_cache()
    def get_todos(self):
        """ Get a list of dicts containing cleaned up todos.

            :param generalpackager.Packager self:
            :rtype: dict[list[str]] """
        return self._get_codeline_todos() + self._get_untested_todos()

    def get_todos_markdown(self, *packagers, drop_package_col=False):
        """ :param generalpackager.Packager self:
            :param drop_package_col: """
        todos = flatten([packager.get_todos() for packager in packagers])
        if drop_package_col:
            todos = [exclusive(todo, "Package") for todo in todos]

        markdown = Markdown(header=self._todo_header)
        if todos:
            markdown.add_table_lines(*todos)
        else:
            markdown.add_node("No todos!")
        return markdown

    def get_description_markdown(self):
        """ :param generalpackager.Packager self: """
        part_of = f"This package and {len(self.get_all()) - 1} other make up {Markdown.link(text='ManderaGeneral', url='https://github.com/ManderaGeneral')}."

        return Markdown(self.localrepo.description, "\n", part_of, header=self.name)

    def get_information_markdown(self, *packagers):
        """ Get information table.

            :param generalpackager.Packager self: """
        if not packagers:
            packagers = (self, )

        markdown = Markdown(header="Information")
        python_url = "https://www.python.org/downloads/release/python-"

        list_of_dicts = []
        for packager in packagers:
            attrs = packager._get_attributes_view().count("\n")
            untested_attrs = len(packager.get_untested_objInfo_dict())
            test_percentage = floor((attrs - untested_attrs) / attrs * 100, 1) if attrs else 100

            list_of_dicts.append({
                "Package": Markdown.link(text=packager.name, url=packager.github.url),
                "Ver": Markdown.link(text=packager.localrepo.version, url=packager.pypi.url),
                "Latest Release": packager.get_latest_release(),
                "Python": ", ".join([Markdown.link(text=ver, url=f"{python_url}{str(ver).replace('.', '')}0/") for ver in packager.python]),
                "Platform": ", ".join(map(str.capitalize, packager.os)),
                "Lvl": packager.get_ordered_index(),
                "Todo": Markdown.link(text=len(packager.get_todos()), url=f"{packager.github.url}#{self._todo_header}"),
                "Tests": f"{test_percentage} %",
            })
        markdown.add_table_lines(*list_of_dicts, sort_by=["Lvl", "Package"])
        return markdown

    def get_installation_markdown(self):
        """ Get install markdown.

            :param generalpackager.Packager self: """
        markdown = Markdown(header="Installation")

        dependencies_required = self.localrepo.install_requires.copy()
        dependencies_optional = list(set().union(*self.localrepo.extras_require.values()))
        dependencies_optional.sort()

        options = {self.name: dependencies_required}
        options.update({f"{self.name}[{key}]": value + dependencies_required for key, value in self.localrepo.extras_require.items()})

        list_of_dicts = []

        for command, packages in options.items():
            row = {"Command": f"`pip install {command}`"}
            for dependency in dependencies_required + dependencies_optional:
                row[Markdown.link(dependency, url=f"https://pypi.org/project/{dependency}", href=True)] = "Yes" if dependency in packages else "No"
            list_of_dicts.append(row)

        markdown.add_table_lines(*list_of_dicts)

        return markdown

    def _configure_contents_markdown(self, markdown):
        """ Configure table of contents lines from markdown.

            :param generalpackager.Packager self:
            :param markdown: """
        parent_markdown = markdown.get_parent(-1, -1)
        markdown.add_pre_lines(parent_markdown.view(custom_repr=lambda md: md.link(md.header, href=True), print_out=False))
        return markdown

    def github_link(self, text, suffix):
        """ Get an HREF link to this repo's github.

            :param generalpackager.Packager self:
            :param text: Text to be clickable.
            :param suffix: URL suffix. """
        url = f"{self.github.url}/{suffix}"
        return Markdown.link(text=text, url=url, href=True)

    def github_link_path_line(self, text, path, line=None):
        """ :param generalpackager.Packager self:
            :param text:
            :param path:
            :param line: """
        if line is None:
            line = 1
        path = Path(path)
        return self.github_link(text=text, suffix=f"blob/{self.commit_sha}/{path.encode()}#L{line}")

    def _attr_repr(self, objInfo):
        """ Return a nice representation of each attribute made by this module, in this case a link to code definition.

            :param generalpackager.Packager self:
            :param generallibrary.ObjInfo objInfo: """
        text = repr(objInfo)
        path = objInfo.file(relative=True)

        line = objInfo.get_definition_line()

        string = self.github_link_path_line(text=text, path=path, line=line)

        if objInfo.name in self.get_untested_objInfo_dict():
            string = f"{string} <b>(Untested)</b>"

        return string

    def _get_attributes_view(self):
        """ :param generalpackager.Packager self: """
        if self.localmodule.module is None:
            return "No module to get attributes"
        return self._cache_get_attributes_view(self.commit_sha)

    @deco_cache()
    def _cache_get_attributes_view(self, commit_sha):
        """ Additional method here to store seperate cache for different commit_shas

            :param generalpackager.Packager self:
            :param commit_sha: """
        return self.localmodule.objInfo.view(custom_repr=self._attr_repr, print_out=False)

    def get_attributes_markdown(self):
        """ Get a recursive view of attributes markdown.

            :param generalpackager.Packager self: """
        return Markdown(header="Attributes").add_pre_lines(self._get_attributes_view())

    def get_footnote_markdown(self, commit=True):
        """ Get a markdown for footnote containing date, time and commit link.

            :param generalpackager.Packager self:
            :param commit: """
        line = f"Generated {Date.now()}"
        if commit:
            line += f" for commit {self.github_link(text=self.commit_sha, suffix=f'commit/{self.commit_sha}')}."

        return Markdown(line).wrap_with_tags("sup")






from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    """ Skipped:
        workflow_unittest
        workflow_sync
        if_publish_bump
        if_publish_publish
        """
    def test_get_triggers(self):
        self.assertIn("branches", Packager().get_triggers())

    def test_step_setup_python(self):
        self.assertIn("3.8", Packager().step_setup_python("3.8"))

    def test_step_install_necessities(self):
        self.assertIn("pip install", Packager().step_install_necessities())

    def test_step_install_package_pip(self):
        self.assertIn("pip install", Packager().step_install_package_pip(*Packager().get_ordered_packagers()))

    def test_step_install_package_git(self):
        self.assertIn("pip install git", Packager().step_install_package_git(*Packager().get_ordered_packagers()))

    def test_get_env(self):
        self.assertIn("TWINE", Packager().get_env())

    def test_steps_setup(self):
        self.assertIn("pip install", Packager().steps_setup("3.8"))

    def test_get_unittest_job(self):
        self.assertIn("pip install", Packager().get_unittest_job())

    def test_get_sync_job(self):
        self.assertIn("pip install", Packager().get_sync_job())

    def test_step_run_packager_method(self):
        self.assertIn("Packager(", Packager().step_run_packager_method("foo"))

    def test_run_ordered_methods(self):
        x = []
        def a(_): x.append(1)
        def b(_): x.append(2)
        Packager().run_ordered_methods(a, b)
        length = len(Packager().get_all())
        self.assertEqual([1] * length + [2] * length, x)





from generalpackager.api.github import GitHub

import unittest


class TestGitHub(unittest.TestCase):
    """ Skipped tests:
        set_website()
        set_topics()
        set_description()
    """
    def test_exists(self):
        self.assertEqual(True, GitHub("generalpackager").exists())

    def test_topics(self):
        github = GitHub("generalpackager")
        self.assertTrue(github.get_topics())

    def test_get_owners_packages(self):
        github = GitHub()
        self.assertEqual(set(), {"generallibrary", "generalfile", "generalvector", "generalpackager"}.difference(github.get_owners_packages()))

        github = GitHub(owner="pandas-dev")
        self.assertEqual(True, "pandas" in github.get_owners_packages())

    def test_get_website(self):
        github = GitHub("generalpackager")
        self.assertEqual(True, "pypi" in github.get_website())

    def test_get_description(self):
        github = GitHub("generalpackager")
        self.assertEqual(True, len(github.get_description()) > 5)






from generallibrary import remove_duplicates, deco_cache


class _PackagerRelations:
    def get_dependencies(self):
        """ Get a list of dependencies as Packagers.
            Combines localmodules dependencies with localrepos install_requires.

            :param generalpackager.Packager self: """
        packagers = {type(self)(localmodule.name) for localmodule in self.localmodule.get_dependencies()}
        packagers.update({type(self)(name) for name in self.localrepo.install_requires})
        return list(packagers)

    def get_dependants(self):
        """ Get a list of dependants as Packagers.
            Same as localmodules but Packager instead of localmodule.

            :param generalpackager.Packager self: """
        packagers = {type(self)(localmodule.name) for localmodule in self.localmodule.get_dependants()}
        return list(packagers)

    def get_ordered_packagers(self):
        """ Get a list of enabled ordered packagers from the dependency chain, sorted by name in each lvl.

            :param generalpackager.Packager self:
            :rtype: list[generalpackager.Packager] """
        return remove_duplicates([packager for packager_set in self.get_ordered(flat=False) for packager in sorted(packager_set, key=lambda x: x.name)])

    def get_owners_package_names(self):
        """ Return a set of owner's packages with intersecting PyPI and GitHub, ignores enabled flag.

            :param generalpackager.Packager self: """
        return self.pypi.get_owners_packages().intersection(self.github.get_owners_packages())

    def general_bumped_set(self):
        """ Return a set of general packagers that have been bumped.

            :param generalpackager.Packager self: """
        return {packager for packager in self.get_all() if packager.is_bumped()}

    def general_changed_dict(self, aesthetic=None):
        """ Return a dict of general packagers with changed files comparing to github.

            :param generalpackager.Packager self:
            :param aesthetic: """
        return {packager: files for packager in self.get_all() if (files := packager.compare_local_to_github(aesthetic=aesthetic))}

    @deco_cache()
    def get_untested_objInfo_dict(self):
        """ :param generalpackager.Packager self:
            :rtype: dict[generallibrary.ObjInfo] """
        if not self.localmodule.objInfo:
            return {}

        filt = lambda objInfo: not self.localrepo.text_in_tests(text=objInfo.name)
        all_objInfo = self.localmodule.objInfo.get_all(filt=filt, traverse_excluded=True)
        return {objInfo.name: objInfo for objInfo in all_objInfo}




from generalpackager.api.shared import _SharedAPI
from generallibrary import Ver, Date, Recycle, get
from generalfile import Path

import requests
import re


def download(url, path):
    """ Todo: Move download to it's own package. """
    data = requests.get(url)
    if data.status_code != 200:
        raise AttributeError(f"Request for url {url} did not yield a status code of 200.'")

    path = Path(path)

    with path.lock():
        path.get_parent().create_folder()
        with open(str(path), "wb") as file:
            file.write(data.content)
    return path


class PyPI(Recycle, _SharedAPI):
    """ Tools to interface pypi.org """
    _recycle_keys = {"name": str, "owner": str}

    def __init__(self, name=None, owner=None):
        if name is None:
            name = "generalpackager"
        if owner is None:
            owner = "Mandera"
        self.name = name
        self.owner = owner

        self.url = f"https://pypi.org/project/{self.name}/"

    def exists(self):
        """ Return whether this API's target exists. """
        return requests.get(url=self.url).status_code == 200

    def get_tarball_url(self, version=None):
        """ Get URL to download tarball. """
        if version is None:
            version = self.get_version()
        return f"https://pypi.io/packages/source/{self.name[0]}/{self.name}/{self.name}-{version}.tar.gz"

    def download(self, path, version=None, overwrite=False):
        """ Download tar ball to cache, extract it, remove tar ball.
            Returns target folder tarball is extracted in. """
        if version is None:
            version = self.get_version()
        path = Path(path)
        temp = Path.get_cache_dir() / "Python/temp.tar.gz"
        download(self.get_tarball_url(version=version), path=temp)
        temp.unpack(base=path, overwrite=overwrite)
        temp.delete(error=False)
        return path / f"{self.name}-{version}"

    def get_owners_packages(self):
        """ Get a set of a owner's packages' names on PyPI. """
        return set(re.findall("/project/(.*)/", requests.get(f"https://pypi.org/user/{self.owner}/").text))

    # Todo: Find a faster fetch for latest PyPI version and datetime.
    def get_version(self):
        """ Get version of latest publish on PyPI. """
        version = get(re.findall(f"{self.name} ([.0-9]+)\n", requests.get(self.url).text), 0)
        return Ver(version) if version else None

    def get_date(self):
        """ Get datetime of latest release. """
        date = get(re.findall('Generated (.+) for commit', requests.get(self.url).text), 0)
        return Date(date) if date else None































from generalpackager import Packager

import unittest


class TestGitHub(unittest.TestCase):
    def test_is_general(self):
        self.assertEqual(True, Packager().is_general())
        self.assertEqual(True, Packager().github.is_general())
        self.assertEqual(True, Packager().pypi.is_general())
        self.assertEqual(True, Packager().localrepo.is_general())
        self.assertEqual(True, Packager().localmodule.is_general())

        self.assertEqual(True, Packager("generalfile").is_general())
        self.assertEqual(True, Packager("generalfile").github.is_general())
        self.assertEqual(True, Packager("generalfile").pypi.is_general())
        self.assertEqual(True, Packager("generalfile").localrepo.is_general())
        self.assertEqual(True, Packager("generalfile").localmodule.is_general())

        self.assertEqual(False, Packager("pandas").is_general())
        self.assertEqual(False, Packager("pandas").github.is_general())
        self.assertEqual(False, Packager("pandas").pypi.is_general())
        self.assertEqual(False, Packager("pandas").localrepo.is_general())
        self.assertEqual(False, Packager("pandas").localmodule.is_general())







from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    """ Skip:
        GenerateFile
        generate
        generate_localfiles
        create_blank
        """
    def test_relative_path_is_aesthetic(self):
        packager = Packager()
        self.assertEqual(False, packager.relative_path_is_aesthetic("setup.py"))
        self.assertEqual(True, packager.relative_path_is_aesthetic("README.md"))
        self.assertEqual(True, packager.relative_path_is_aesthetic(packager.localrepo.get_readme_path()))

    def test_filter_relative_filenames(self):
        packager = Packager()
        self.assertEqual(["setup.py"], packager.filter_relative_filenames("setup.py", aesthetic=None))
        self.assertEqual(["setup.py"], packager.filter_relative_filenames("setup.py", aesthetic=False))
        self.assertEqual([], packager.filter_relative_filenames("setup.py", aesthetic=True))

    def test_compare_local_to_github(self):
        setup_workdir(use_working_dir=True)
        packager = Packager()
        packager.compare_local_to_github()

    def test_compare_local_to_pypi(self):
        setup_workdir(use_working_dir=True)
        packager = Packager()
        packager.compare_local_to_pypi()

    def test_generate_setup(self):
        packager = Packager()
        text = packager.generate_setup()
        self.assertIn(str(packager.localrepo.version), text)
        self.assertIn(str(packager.localrepo.name), text)

    def test_generate_manifest(self):
        packager = Packager()
        text = packager.generate_manifest()
        self.assertIn("include metadata.json", text)

    def test_generate_git_exclude(self):
        packager = Packager()
        text = packager.generate_git_exclude()
        self.assertIn(".idea", text)

    def test_generate_license(self):
        packager = Packager()
        text = packager.generate_license()
        self.assertIn("Mandera", text)

    def test_generate_workflow(self):
        packager = Packager()
        text = packager.generate_workflow()
        self.assertIn("runs-on", text)

    def test_generate_readme(self):
        packager = Packager()
        text = str(packager.generate_readme())
        self.assertIn("pip install", text)

    def test_generate_personal_readme(self):
        packager = Packager()
        text = str(packager.generate_personal_readme())
        self.assertIn("generallibrary", text)

    def test_get_new_packager(self):
        a = Packager()
        b = a.get_new_packager()
        self.assertEqual(a.name, b.name)
        self.assertIsNot(a, b)
        self.assertIs(b, Packager())






from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    """ Skipped:
        sync_github_metadata
        commit_and_push """
    # def test_clone_repo(self):
    #     setup_workdir(use_working_dir=True)
    #     self.assertEqual(True, Packager().clone_repo("").exists())






from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    def test_badges_dict(self):
        self.assertLess(2, len(Packager().get_badges_dict()))

    def test_get_todos(self):
        Packager().get_todos()

    def test_get_todos_markdown(self):
        Packager().get_todos_markdown()

    def test_get_description_markdown(self):
        self.assertIn("generalpackager", Packager().get_description_markdown())

    def test_get_information_markdown(self):
        self.assertIn("generalpackager", Packager().get_information_markdown())

    def test_get_installation_markdown(self):
        self.assertIn("pip install generalpackager", Packager().get_installation_markdown())

    def test_github_link(self):
        self.assertIn("generalpackager", Packager().github_link("foo", "bar"))

    def test_github_link_path_line(self):
        self.assertIn("generalpackager", Packager().github_link_path_line("foo", "bar"))

    def test_get_attributes_markdown(self):
        self.assertIn("generalpackager", Packager().get_attributes_markdown())

    def test_get_footnote_markdown(self):
        self.assertIn("generalpackager", Packager().get_footnote_markdown())







class _SharedAPI:
    def is_general(self):
        """ Return whether name is general or not.

            :param generalpackager.api.pypi.PyPI or generalpackager.api.local_module.LocalModule or generalpackager.api.local_repo.LocalRepo or generalpackager.api.github.GitHub self: """
        return self.name.startswith("general") or self.name in ("manderageneral", )





from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    def test_get_topics(self):
        self.assertIn("windows", Packager().get_topics())

    def test_get_classifiers(self):
        self.assertIn("Operating System :: Microsoft :: Windows", Packager().get_classifiers())

    def test_is_bumped(self):
        packager = Packager()
        packager.is_bumped()
        version = packager.localrepo.version
        packager.localrepo.bump_version()
        self.assertEqual(True, packager.is_bumped())
        packager.localrepo.version = version





from generallibrary import EnvVar

PACKAGER_GITHUB_API = EnvVar("PACKAGER_GITHUB_API", "secrets.PACKAGER_GITHUB_API")
TWINE_USERNAME = EnvVar("TWINE_USERNAME", "secrets.TWINE_USERNAME")
TWINE_PASSWORD = EnvVar("TWINE_PASSWORD", "secrets.TWINE_PASSWORD")

from generalpackager.packager import Packager
from generalpackager.api.local_repo import LocalRepo
from generalpackager.api.local_module import LocalModule
from generalpackager.api.github import GitHub
from generalpackager.api.pypi import PyPI






from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    def test_get_ordered_packagers(self):
        self.assertLess(3, len(Packager().get_ordered_packagers()))

    def test_get_owners_package_names(self):
        self.assertLess(4, len(Packager().get_owners_package_names()))

    def test_general_bumped_set(self):
        packager = Packager()
        packager.general_bumped_set()

        version = packager.localrepo.version
        packager.localrepo.bump_version()
        self.assertLess(0, len(packager.general_bumped_set()))
        packager.localrepo.version = version

    def test_general_changed_dict(self):
        Packager().general_changed_dict()

    def test_get_untested_objInfo_dict(self):
        self.assertIsInstance(Packager().get_untested_objInfo_dict(), dict)





class _PackagerMetadata:
    def _topics_to_classifiers(self, *topics):
        """ :param generalpackager.Packager self: """
        classifiers = []
        for topic in topics:
            if topic.startswith("python"):
                major = topic[6]
                minor = topic[7:]
                classifiers.append(f"Programming Language :: Python :: {major}.{minor}")
            else:
                classifiers.append(self._lib[topic])
        return classifiers

    def get_topics(self):
        """ Get a complete list of topics by using package specific as well as hardcoded magic values.

            :param generalpackager.Packager self: """
        topics = self.localrepo.topics.copy()
        topics.extend([f"python{ver.replace('.', '')}" for ver in self.python])
        topics.append(f"{self.license}-license")
        topics.extend(self.os)
        return topics

    def get_classifiers(self):
        """ Get a complete list of classifiers generated from topics and other metadata.

            :param generalpackager.Packager self: """
        return self._topics_to_classifiers(*self.get_topics())

    def is_bumped(self):
        """ Return whether this package has been bumped by comparing PyPI and LocalRepo's versions.

            :param generalpackager.Packager self: """
        return self.pypi.get_version() is None or self.localrepo.version > self.pypi.get_version()

    _lib = {
        "planning": "Development Status :: 1 - Planning",
        "pre-alpha": "Development Status :: 2 - Pre-Alpha",
        "alpha": "Development Status :: 3 - Alpha",
        "beta": "Development Status :: 4 - Beta",
        "production/Stable": "Development Status :: 5 - Production/Stable",
        "mature": "Development Status :: 6 - Mature",
        "inactive": "Development Status :: 7 - Inactive",

        "utility": "Topic :: Utilities",

        "tool": "Topic :: Software Development :: Build Tools",
        "library": "Topic :: Software Development :: Libraries",
        "gui": "Topic :: Software Development :: User Interfaces",

        "file-manager": "Topic :: Desktop Environment :: File Managers",

        "mit-license": "License :: OSI Approved :: MIT License",

        "windows": "Operating System :: Microsoft :: Windows",
        "macos": "Operating System :: MacOS",
        "ubuntu": "Operating System :: POSIX :: Linux",
    }



""" Methods specific for my general packages.
    Isolatable methods are put inside APIs.

    Todo: Prevent workflow using pypi to install a general package. """

from generallibrary import initBases, NetworkDiagram, Recycle
from generalpackager.api.shared import _SharedAPI
from generalpackager.api.local_repo import LocalRepo
from generalpackager.api.local_module import LocalModule
from generalpackager.api.github import GitHub
from generalpackager.api.pypi import PyPI

from generalpackager.packager_files import _PackagerFiles
from generalpackager.packager_github import _PackagerGitHub
from generalpackager.packager_markdown import _PackagerMarkdown
from generalpackager.packager_metadata import _PackagerMetadata
from generalpackager.packager_pypi import _PackagerPypi
from generalpackager.packager_workflow import _PackagerWorkflow
from generalpackager.packager_relations import _PackagerRelations


@initBases
class Packager(Recycle, _SharedAPI, NetworkDiagram, _PackagerMarkdown, _PackagerGitHub, _PackagerFiles, _PackagerMetadata, _PackagerPypi, _PackagerWorkflow, _PackagerRelations):
    """ Uses APIs to manage 'general' package.
        Contains methods that require more than one API as well as methods specific for ManderaGeneral. """

    author = 'Rickard "Mandera" Abraham'
    email = "rickard.abraham@gmail.com"
    license = "mit"
    python = "3.8", "3.9"  # Only supports basic definition with tuple of major.minor
    os = "windows", "ubuntu"  # , "macos"

    git_exclude_lines = ".idea", "build", "dist", "*.egg-info", "__pycache__", ".git", "test/tests", "PKG-INFO", "setup.cfg"
    _recycle_keys = LocalModule._recycle_keys
    _recycle_keys["path"] = str

    summary_packagers = []

    def __init__(self, name=None, github_owner=None, pypi_owner=None, path=None):
        self.localmodule = LocalModule(name=name)
        self.name = self.localmodule.name

        self.localrepo = LocalRepo(name=self.name, path=path)
        self.path = self.localrepo.path

        self.github = GitHub(name=self.name, owner=github_owner)
        self.pypi = PyPI(name=self.name, owner=pypi_owner)

        # Quick fix, should probably put download in a workflow step instead
        # Yeah because this prevents us from creating a new package with create_blank
        # if (self.localmodule.is_general() or self.name in ("Mandera", ".github")) and not self.localrepo.exists():
        #     self.github.download(path=self.path.get_parent())

    def spawn_children(self):
        """ :param generalpackager.Packager self: """
        for packager in self.get_dependants():
            if packager.is_general() and packager.localrepo.enabled:
                packager.set_parent(parent=self)

    def spawn_parents(self):
        """ :param generalpackager.Packager self: """
        for packager in self.get_dependencies():
            if packager.is_general() and packager.localrepo.enabled:
                self.set_parent(parent=packager)

    def exists(self):
        """ Just check GitHub for now. """
        return self.github.exists()

    def __repr__(self):
        return f"<Packager: {self.name}>"

Packager.summary_packagers = [
    Packager(name="Mandera", github_owner="Mandera"),
    Packager(name=".github", github_owner="ManderaGeneral"),
]




















































from generalpackager.api.shared import _SharedAPI
from generalpackager import PACKAGER_GITHUB_API
from generallibrary import Recycle, deco_cache
from generalfile import Path

import requests
import json
import re
from git import Repo


class GitHub(Recycle, _SharedAPI):
    """ Tools to interface a GitHub Repository.
        Todo: Get and Set GitHub repo private. """
    _recycle_keys = {"name": str, "owner": str}

    def __init__(self, name=None, owner=None):
        if name is None:
            name = "generalpackager"
        if owner is None:
            owner = "ManderaGeneral"
        self.name = name
        self.owner = owner

        self.url = f"https://github.com/{self.owner}/{self.name}"

    def exists(self):
        """ Return whether this API's target exists. """
        return requests.get(url=self.url).status_code == 200

    def download(self, path, overwrite=False):
        """ Clone a GitHub repo into a path.
            Creates a folder with Package's name first.
            Target must be empty.

            :param generalpackager.Packager self:
            :param path:
            :param overwrite: """
        if not self.exists():
            return

        path = Path(path) / self.name

        if path.exists():
            if overwrite:
                path.delete()
            else:
                raise AttributeError(f"Clone target exists and overwrite is False.")

        Repo.clone_from(url=self.url, to_path=path)
        return path

    def get_owners_packages(self):
        """ Get a set of a owner's packages' names on GitHub. """
        return set(re.findall(f'"/{self.owner}/(.+)/hovercard"', requests.get(f"https://github.com/{self.owner}?tab=repositories").text))

    def _api_url(self, endpoint=None):
        """ Get URL from owner, name and enpoint. """
        return "/".join(("https://api.github.com", "repos", self.owner, self.name) + ((endpoint, ) if endpoint else ()))

    def get_website(self):
        """ Get website specified in repository details.

            :rtype: list[str] """
        return self._request(method="get").json()["homepage"]

    def set_website(self, website):
        """ Set a website for the GitHub repository. """
        return self._request(method="patch", name=self.name, homepage=website)

    def get_topics(self):
        """ Get a list of topics in the GitHub repository.

            :rtype: list[str] """
        return self._request(method="get", endpoint="topics").json()["names"]

    def set_topics(self, *topics):
        """ Set topics for the GitHub repository.

            :param str topics: """
        return self._request(method="put", endpoint="topics", names=topics)

    def get_description(self):
        """ Get a string of description in the GitHub repository.

            :rtype: list[str] """
        return self._request(method="get").json()["description"]

    def set_description(self, description):
        """ Set a description for the GitHub repository. """
        return self._request(method="patch", name=self.name, description=description)

    @deco_cache()
    def _request(self, method="get", url=None, endpoint=None, **data):
        """ :rtype: requests.Response """
        method = getattr(requests, method.lower())

        kwargs = {
            "headers": {"Accept": "application/vnd.github.mercy-preview+json"},
            "auth": (self.owner, PACKAGER_GITHUB_API.value),
        }
        if data:
            kwargs["data"] = json.dumps(data)
        if url is None:
            url = self._api_url(endpoint=endpoint)
        return method(url=url, **kwargs)



















from generalpackager.api.shared import _SharedAPI
from generalfile import Path
from generallibrary import Ver, deco_cache, Recycle, terminal

from setuptools import find_namespace_packages
import re
from git import Repo
import sys


def load_metadata_before(func):
    def _wrapper(self, *args, **kwargs):
        """ :param LocalRepo self: """
        if not self.has_loaded_metadata:
            self.load_metadata()
        return func(self, *args, **kwargs)
    return _wrapper


class LocalRepo(Recycle, _SharedAPI):
    """ Tools to help Path interface a Local Python Repository.
        Todo: Search for imports to list dependencies. """

    enabled = True
    private = False
    installable = True  # Might not need this, generalmainframe can be installable, just not on pypi
    name = ...
    version = "0.0.1"  # type: Ver
    description = "Missing description."
    install_requires = ["generallibrary"]
    extras_require = {}
    topics = []
    manifest = []

    metadata_keys = [key for key, value in locals().items() if not key.startswith("_")]
    _recycle_keys = {"name": lambda name: str(LocalRepo.get_path_from_name(name=name))}

    def __init__(self, name=None, path=None):
        if path is None:
            self.path = self.get_path_from_name(name=name)
        else:
            path = Path(path)
            if not path.endswith(name):
                path /= name
            self.path = path
        self.has_loaded_metadata = False

    def __repr__(self):
        return str(self.path)

    def has_metadata(self):
        return self.get_metadata_path().exists()

    def load_metadata(self):
        """ Read metadata path and load values. """
        self.has_loaded_metadata = True

        if self.get_metadata_path().exists():
            for key, value in self.get_metadata_path().read().items():
                setattr(self, f"_{key}", value)

            if self.extras_require:
                self.extras_require["full"] = list(set().union(*self.extras_require.values()))
                self.extras_require["full"].sort()

            self.version = Ver(self.version)

        if self.name is Ellipsis:
            self.name = self.path.name()

    def get_metadata_dict(self):
        """ Get current metadata values as a dict. """
        return {key: str(getattr(self, key)) if key == "version" else getattr(self, key) for key in self.metadata_keys}

    def write_metadata(self):
        """ Write to metadata path using current metadata values. """
        self.get_metadata_path().write(self.get_metadata_dict(), overwrite=True, indent=4)

    def exists(self):
        """ Return whether this API's target exists. """
        return self.path_exists(path=self.path)

    @classmethod
    def path_exists(cls, path):
        if path.is_file() or not path.exists():
            return False
        return bool(path.get_child(filt=lambda x: x.name() in ("README.md", ), traverse_excluded=True))
        # return bool(path.get_child(filt=lambda x: x.name() in ("setup.py", ), traverse_excluded=True))  # setup.py was not included in pypi's sdist

    @classmethod
    def get_repo_path_parent(cls, path=None):
        """ Iterate self and parents to return first path where LocalRepo can be created or None.

            :param Path or any path:
            :rtype: Path """
        return Path(path=path).absolute().get_parent(depth=-1, include_self=True, filt=LocalRepo.path_exists, traverse_excluded=True)

    @classmethod
    def get_repo_path_child(cls, folder_path):
        """ Return whether there's atleast one repo in folder. """
        return Path(folder_path).get_child(filt=LocalRepo.path_exists, traverse_excluded=True)

    @classmethod
    def get_repos_path(cls, path=None):
        """ Try to return absolute path pointing to folder containing repos or None.

            :rtype: Path """
        filt = lambda path: LocalRepo.get_repo_path_child(path) and not path.match("/test")
        return Path(path).absolute().get_parent(depth=-1, include_self=True, filt=filt, traverse_excluded=True)

    @classmethod
    def get_path_from_name(cls, name=None):
        """ Get path from name.

            :param Path or any name:
            :rtype: Path """
        # Check if inside correct repo already
        path = cls.get_repo_path_parent()
        if path and (name is None or name == path.name()):
            return path

        # if Path.get_working_dir().name() == name:
        #     return Path.get_working_dir()

        if name is None:
            name = "generalpackager"

        # Check if there's a parent that has a repo child that exists
        repos_path = cls.get_repos_path()
        if repos_path:
            return repos_path / name

        return Path(name).absolute()

    @load_metadata_before
    def _metadata_getter(self, key):
        return getattr(self, f"_{key}")

    @load_metadata_before
    def _metadata_setter(self, key, value):
        """ Set a metadata's key both in instance and json file. """
        if self.has_metadata() and value != getattr(self, f"_{key}", ...):
            metadata = self.get_metadata_path().read()
            metadata[key] = str(value) if key == "version" else value  # Todo: Decoupled JSON serialize instructions with custom dumps in lib.
            self.get_metadata_path().write(metadata, overwrite=True, indent=4)
        setattr(self, f"_{key}", value)

    def get_readme_path(self):              return self.path / "README.md"
    def get_org_readme_path(self):          return self.path / "profile/README.md"
    def get_metadata_path(self):            return self.path / "metadata.json"
    def get_git_exclude_path(self):         return self.path / ".git/info/exclude"
    def get_setup_path(self):               return self.path / "setup.py"
    def get_manifest_path(self):            return self.path / "MANIFEST.in"
    def get_license_path(self):             return self.path / "LICENSE"
    def get_workflow_path(self):            return self.path / ".github/workflows/workflow.yml"
    def get_test_path(self):                return self.path / f"{self.name}/test"
    def get_test_template_path(self):       return self.get_test_path() / f"test_{self.name}.py"
    def get_init_path(self):                return self.path / self.name / "__init__.py"
    def get_randomtesting_path(self):       return self.path / "randomtesting.py"
    def get_generate_path(self):            return self.path / "generate.py"

    @deco_cache()
    def get_test_paths(self):
        """ Yield paths to each test python file. """
        return self.get_test_path().get_children(depth=-1, filt=lambda path: path.endswith(".py") and not path.match("/tests/"), traverse_excluded=True)

    @deco_cache()
    def text_in_tests(self, text):
        """ Return whether text exists in one of the test files. """
        for path in self.get_test_paths():
            if path.contains(text=str(text)):
                return True
        return False

    def get_package_paths_gen(self):
        """ Yield Paths pointing to each folder containing a Python file in this local repo, aka `namespace package`.
            Doesn't include self.path """
        for package in find_namespace_packages(where=str(self.path)):
            path = self.path / package.replace(".", "/")
            if not path.match("/dist", "/build"):
                yield path

    def git_changed_files(self):
        """ Get a list of changed files using local .git folder. """
        repo = Repo(str(self.path))
        return [Path(file) for file in re.findall("diff --git a/(.*) " + "b/", repo.git.diff())]

    def bump_version(self):
        """ Bump micro version in metadata.json. """
        self.version = self.version.bump()

    def pip_install(self):
        """ Install this repository with pip, WITHOUT -e flag.
            Subprocess messed up -e flag compared to doing it in terminal, so use the normal one."""
        with self.path.as_working_dir():
            terminal("pip", "install", "-e", ".")

    def pip_uninstall(self):
        """ Uninstall this repository with pip."""
        terminal("-m", "pip", "uninstall", "-y", self.name, python=True)

    def unittest(self):
        """ Run unittests for this repository. """
        terminal("-m", "unittest", "discover", str(self.get_test_path()), python=True)

    def create_sdist(self):
        """ Create source distribution. """
        with self.path.as_working_dir():
            terminal("setup.py", "sdist", "bdist_wheel", python=True)

    def upload(self):
        """ Upload local repo to PyPI.
            Todo: Make sure twine is installed when trying to upload to pypi. """
        if self.private:
            raise AttributeError("Cannot upload private repo.")

        self.create_sdist()
        with self.path.as_working_dir():
            terminal("-m", "twine", "upload", "dist/*", python=True)

    def generate_exe(self, file_path, suppress=False):
        """ Generate an exe file for target file_path python file. """
        with self.path.as_working_dir():
            terminal("-m", "PyInstaller", file_path, "--onefile", "--windowed", python=True, suppress=suppress)
            # terminal("-m", "PyInstaller", file_path, "--onefile", "--windowed", "--name", self.name, python=True, suppress=suppress)  # Failed for some reason

for key in LocalRepo.metadata_keys:
    value = getattr(LocalRepo, key)
    setattr(LocalRepo, f"_{key}", value)
    setattr(LocalRepo, key, property(
        fget=lambda self, key=key: LocalRepo._metadata_getter(self, key),
        fset=lambda self, value, key=key: LocalRepo._metadata_setter(self, key, value),
    ))






































from generallibrary import CodeLine, Markdown, Date, exclusive, deco_cache, cache_clear, Timer
from generalfile import Path


class GenerateFile:
    """ Handle generation of files. """
    def __init__(self, path, text_func, packager, aesthetic, overwrite=True):
        """ :param generalpackager.Packager packager: """
        self.text_func = text_func
        self.packager = packager
        self.aesthetic = aesthetic
        self.overwrite = overwrite

        self.relative_path = path.relative(base=packager.path)
        self.path = packager.path / self.relative_path

    def generate(self):
        """ Generate actual file. """
        if self.overwrite or not self.path.exists():
            self.path.text.write(f"{self.text_func()}\n", overwrite=self.overwrite)

    def __str__(self):
        return f"<GenerateFile: {self.packager.name} - {self.relative_path}>"


class _PackagerFiles:
    """ Generates setup, license and gitexclude.
        Only changed non-aesthetic files can trigger a version bump and publish. """
    extra_aesthetic = "randomtesting.py",  # "licenses"
    extra_non_aesthetic = tuple()

    _todo_header = "Todo"

    def __init_post__(self):
        """ Todo: Watermark generated files to prevent mistake of thinking you can modify them directly.

            :param generalpackager.Packager self: """
        self.file_setup =           GenerateFile(self.localrepo.get_setup_path(), self.generate_setup, self, aesthetic=False)
        self.file_manifest =        GenerateFile(self.localrepo.get_manifest_path(), self.generate_manifest, self, aesthetic=False)

        self.file_git_exclude =     GenerateFile(self.localrepo.get_git_exclude_path(), self.generate_git_exclude, self, aesthetic=True)
        self.file_license =         GenerateFile(self.localrepo.get_license_path(), self.generate_license, self, aesthetic=True)
        self.file_workflow =        GenerateFile(self.localrepo.get_workflow_path(), self.generate_workflow, self, aesthetic=True)
        self.file_readme =          GenerateFile(self.localrepo.get_readme_path(), self.generate_readme, self, aesthetic=True)
        self.file_generate =        GenerateFile(self.localrepo.get_generate_path(), self.generate_generate, self, aesthetic=True)

        self.file_init =            GenerateFile(self.localrepo.get_init_path(), self.generate_init, self, aesthetic=False, overwrite=False)
        self.file_randomtesting =   GenerateFile(self.localrepo.get_randomtesting_path(), self.generate_randomtesting, self, aesthetic=True, overwrite=False)
        self.file_test_template =   GenerateFile(self.localrepo.get_test_template_path(), self.generate_test_template, self, aesthetic=False, overwrite=False)

        self.files = [getattr(self, key) for key in dir(self) if key.startswith("file_")]  # type: list[GenerateFile]
        self.files_by_relative_path = {file.relative_path: file for file in self.files}

        # Organization secret name is .github, user secret name is user
        secret_readme_path = self.localrepo.get_org_readme_path() if self.name == ".github" else self.localrepo.get_readme_path()
        self.file_secret_readme = GenerateFile(secret_readme_path, self.generate_personal_readme, self, aesthetic=True)

    def get_new_packager(self):
        """ :param generalpackager.Packager self: """
        self.recycle_clear()
        self.localrepo.recycle_clear()
        self.localmodule.recycle_clear()
        self.github.recycle_clear()
        self.pypi.recycle_clear()
        return type(self)(self.name)

    def create_blank_locally(self, install=True):
        """ Create a new general package locally only.

            :param generalpackager.Packager self:
            :param install: Whether to pip install. """
        self.localrepo.write_metadata()
        self.generate_localfiles()
        if install:
            self.localrepo.pip_install()

        new_self = self.get_new_packager()  # Reset caches to get updated files
        new_self.generate_localfiles()

    def relative_path_is_aesthetic(self, relative_path):
        """ Relative to package path. False if not defined as a GenerateFile instance.

            :param generalpackager.Packager self:
            :param Path or str relative_path: """
        relative_path = Path(relative_path).relative(self.path)
        aesthetic_attr = getattr(self.files_by_relative_path.get(relative_path, None), "aesthetic", None)
        if aesthetic_attr is None:
            if relative_path.match(*self.extra_aesthetic):
                return True
            # elif relative_path.match(*self.extra_non_aesthetic):
            #     return False
            else:
                return False
        return aesthetic_attr

    def filter_relative_filenames(self, *filenames, aesthetic):
        """ If aesthetic is None then it doesn't filter any.
            True will return only aesthetic.
            False will return only non-aesthetic.

            :param generalpackager.Packager self:
            :param bool or None aesthetic: """
        return [path for path in filenames if aesthetic is None or aesthetic is self.relative_path_is_aesthetic(path)]

    @deco_cache()
    def _compare_local(self, platform, aesthetic):
        """ :param generalpackager.Packager self: """
        unpack_target = Path.get_cache_dir() / "Python"
        package_path = platform.download(path=unpack_target, overwrite=True)

        filt = lambda path: not path.match(*self.git_exclude_lines)

        differing_files = self.path.get_differing_files(target=package_path, filt=filt)

        return self.filter_relative_filenames(*differing_files, aesthetic=aesthetic)

    def compare_local_to_github(self, aesthetic=None):
        """ Get a list of changed files compared to remote with optional aesthetic files.

            :param generalpackager.Packager self:
            :param aesthetic: """
        return self._compare_local(platform=self.github, aesthetic=aesthetic)

    def compare_local_to_pypi(self, aesthetic=None):
        """ Get a list of changed files compared to pypi with optional aesthetic files.

            :param generalpackager.Packager self:
            :param aesthetic: """
        return self._compare_local(platform=self.pypi, aesthetic=aesthetic)

    def generate_setup(self):
        """ Generate setup.py.

            :param generalpackager.Packager self: """
        readme_path = self.localrepo.get_readme_path().relative(self.localrepo.get_setup_path().get_parent())
        last_version_split = self.python[-1].split(".")
        last_version_bumped_micro = f"{last_version_split[0]}.{int(last_version_split[1]) + 1}"
        setup_kwargs = {
            "name": f'"{self.localrepo.name}"',
            "author": f"'{self.author}'",
            "author_email": f'"{self.email}"',
            "version": f'"{self.localrepo.version}"',
            "description": f'"{self.localrepo.description}"',
            "long_description": "long_description",
            "long_description_content_type": '"text/markdown"',
            "install_requires": self.localrepo.install_requires,
            "url": f'"{self.github.url}"',
            "license": f'"{self.license}"',
            "python_requires": f'">={self.python[0]}, <{last_version_bumped_micro}"',
            "packages": 'find_namespace_packages(exclude=("build*", "dist*"))',
            "extras_require": self.localrepo.extras_require,
            "classifiers": self.get_classifiers(),
            # "include_package_data": True,
        }

        top = CodeLine()
        top.add_node(CodeLine("from setuptools import setup, find_namespace_packages", space_before=1))
        top.add_node(CodeLine("from pathlib import Path", space_after=1))

        top.add_node(CodeLine("try:")).add_node(CodeLine("long_description = (Path(__file__).parent / 'README.md').read_text(encoding='utf-8')"))
        top.add_node(CodeLine("except FileNotFoundError:")).add_node(CodeLine("long_description = 'Readme missing'", space_after=1))

        setup = top.add_node(CodeLine("setup("))
        for key, value in setup_kwargs.items():
            if isinstance(value, list) and value:
                list_ = setup.add_node(CodeLine(f"{key}=["))
                for item in value:
                    list_.add_node(CodeLine(f"'{item}',"))
                setup.add_node(CodeLine("],"))
            elif isinstance(value, dict) and value:
                dict_ = setup.add_node(CodeLine(f"{key}={{"))
                for k, v in value.items():
                    dict_.add_node(CodeLine(f"'{k}': {v},"))
                setup.add_node(CodeLine("},"))
            else:
                setup.add_node(CodeLine(f"{key}={value},"))

        top.add_node(CodeLine(")"))

        return top.text()

    def generate_manifest(self):
        """ Generate manifest file.

            :param generalpackager.Packager self: """
        default_manifest = [
            self.localrepo.get_metadata_path().relative(self.path),
        ]
        return "\n".join([f"include {path}" for path in self.localrepo.manifest + default_manifest])

    def generate_git_exclude(self):
        """ Generate git exclude file.

            :param generalpackager.Packager self: """
        return "\n".join(self.git_exclude_lines)

    def generate_license(self):
        """ Generate LICENSE by using Packager.license.

            :param generalpackager.Packager self: """
        text = Path(self.localrepo.get_repos_path() / f"generalpackager/generalpackager/licenses/{self.license}").text.read()
        assert "$" in text
        text = text.replace("$year", str(Date.now().datetime.year))
        text = text.replace("$author", self.author)
        assert "$" not in text

        return text

    def generate_workflow(self):
        """ Generate workflow.yml.

            :param generalpackager.Packager self: """
        workflow = CodeLine()
        workflow.indent_str = " " * 2

        workflow.add_node("name: workflow")
        workflow.add_node(self.get_triggers())

        jobs = workflow.add_node("jobs:")
        jobs.add_node(self.get_unittest_job())
        jobs.add_node(self.get_sync_job())

        return workflow.text()

    def generate_readme(self):
        """ Generate readme markdown and overwrite README.md in local repo.

            :param generalpackager.Packager self: """
        # Description
        markdown = self.get_description_markdown()

        # Information
        self.get_information_markdown().set_parent(parent=markdown)

        # Table of contents
        contents = Markdown(header="Contents", parent=markdown)

        # Installation
        self.get_installation_markdown().set_parent(parent=markdown)

        # Attributes
        self.get_attributes_markdown().set_parent(parent=markdown)

        # Todos
        self.get_todos_markdown(self, drop_package_col=True).set_parent(parent=markdown)

        # Table of contents - Configuration
        self._configure_contents_markdown(markdown=contents)

        # Generation timestamp
        self.get_footnote_markdown().set_parent(parent=markdown)

        return markdown

    def generate_personal_readme(self):
        """ Generate personal readme markdown and overwrite README.md in local repo.

            :param generalpackager.Packager self: """
        ordered_packagers = type(self)().get_ordered_packagers()

        # Description
        markdown = Markdown(header="ManderaGeneral").add_list_lines(
            "A collection of connected packages.",
            "Violently updated with little regard for backwards compatability.",
            "Automatic workflows to unittest, sync and publish.",
        )

        # Package information
        self.get_information_markdown(*ordered_packagers).set_parent(parent=markdown)

        # Todos
        self.get_todos_markdown(*ordered_packagers).set_parent(parent=markdown)

        # Generation timestamp
        self.get_footnote_markdown(commit=False).set_parent(parent=markdown)

        return markdown

    def generate_init(self):
        """ Generate __init__.py.

            :param generalpackager.Packager self: """
        codeline = CodeLine(f"", space_before=1, space_after=50)

        return codeline

    def generate_randomtesting(self):
        """ Generate randomtesting.py.

            :param generalpackager.Packager self: """
        codeline = CodeLine(f"from {self.name} import *", space_before=1, space_after=50)

        return codeline

    def generate_generate(self):
        """ Generate randomtesting.py.

            :param generalpackager.Packager self: """
        top = CodeLine()
        top.add_node(CodeLine(f"from generalpackager import Packager", space_before=1, space_after=1))
        top.add_node(CodeLine(f"""Packager("{self.name}").generate_localfiles(print_out=True)""", space_after=50))

        return top

    def generate_test_template(self):
        """ Generate test template.

            :param generalpackager.Packager self: """
        top = CodeLine()
        top.add_node(CodeLine("from unittest import TestCase", space_after=2))
        top.add_node("class Test(TestCase):").add_node("def test(self):").add_node("pass")

        return top

    def generate_localfiles(self, aesthetic=True, print_out=False):
        """ Generate all local files.

            :param aesthetic:
            :param generalpackager.Packager self:
            :param print_out: """
        timer = Timer()

        # Not in files because it writes with json not text, it's also a bit unique
        self.localrepo.write_metadata()

        for generate in self.files:
            if aesthetic or not generate.aesthetic:
                if print_out:
                    print(f"Generating {generate}")
                generate.generate()
        if print_out:
            timer.print()






























from generalpackager import PACKAGER_GITHUB_API

from git import Repo


class _PackagerGitHub:
    """ Sync metadata. """
    def __init__(self):
        self.commit_sha = "master"

    def sync_github_metadata(self):
        """ Sync GitHub with local metadata.

            :param generalpackager.Packager self: """
        self.github.set_website(self.pypi.url)
        self.github.set_description(self.localrepo.description)
        self.github.set_topics(*self.get_topics())

    def commit_and_push(self, message=None, tag=False):
        """ Commit and push this local repo to GitHub.
            Return short sha1 of pushed commit.

            :param generalpackager.Packager self:
            :param message:
            :param tag: """
        repo = Repo(str(self.path))
        repo.git.add(A=True)
        repo.index.commit(message=str(message) or "Automatic commit.")
        remote = repo.remote()
        remote.set_url(f"https://Mandera:{PACKAGER_GITHUB_API}@github.com/{self.github.owner}/{self.name}.git")

        if tag:
            tag_ref = repo.create_tag(f"v{self.localrepo.version}", force=True)
            remote.push(refspec=tag_ref)
        try:
            self.commit_sha = remote.push()[0].summary.split("..")[1].rstrip()
        except OSError:  # Just suppressing weird invalid handle error
            pass
        return self.commit_sha

















from generalpackager.api.local_module import LocalModule

import unittest


class TestLocalModule(unittest.TestCase):
    """ Skipped tests:
    """
    def test_exists(self):
        self.assertEqual(True, LocalModule("generalpackager").exists())
        self.assertEqual(False, LocalModule("doesntexist").exists())

    def test_module(self):
        self.assertEqual("generalpackager", LocalModule().module.__name__)

    def test_objInfo(self):
        self.assertLess(10, len(LocalModule().objInfo.get_children(depth=-1)))

    def test_get_env_vars(self):
        self.assertGreater(len(LocalModule().get_env_vars()), 1)

    def test_get_all_local_modules(self):
        self.assertIn(LocalModule("generallibrary"), LocalModule().get_all_local_modules())
        self.assertIn(LocalModule("generalfile"), LocalModule().get_all_local_modules())
        self.assertNotIn(LocalModule("doesntexist"), LocalModule().get_all_local_modules())

    def test_get_dependencies(self):
        self.assertIn(LocalModule("generallibrary"), LocalModule().get_dependencies())
        self.assertIn(LocalModule("generalfile"), LocalModule().get_dependencies())
        self.assertNotIn(LocalModule("doesntexist"), LocalModule().get_dependencies())

    def test_get_dependants(self):
        self.assertEqual([], LocalModule().get_dependants())












from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    def test_get_latest_release(self):
        self.assertIn("CE", str(Packager().get_latest_release()))






from generalpackager.api.pypi import PyPI
from generalfile.test.setup_workdir import setup_workdir
from generallibrary import Date

import unittest


class TestPyPI(unittest.TestCase):
    def test_exists(self):
        self.assertEqual(True, PyPI("generalpackager").exists())
        self.assertEqual(False, PyPI("random-package_that,cant.exist").exists())

    def test_get_tarball_url(self):
        pypi = PyPI("generalpackager")
        self.assertEqual(True, pypi.name in pypi.get_tarball_url())
        self.assertEqual(True, pypi.name in pypi.get_tarball_url(version="1.0.0"))

    def test_download(self):
        setup_workdir(use_working_dir=True)

        path = PyPI("generalpackager").download(path="repo")
        self.assertEqual(True, "generalpackager" in path.get_child())

        with self.assertRaises(AttributeError):
            PyPI("generalpackager").download(path="repo", version="0.0.111")

        path = PyPI("generalpackager").download(path="repo", version="0.0.11", overwrite=True)
        self.assertEqual(2, len(path.get_parent().get_children()))

    def test_get_owners_packages(self):
        github = PyPI()
        self.assertEqual(set(), {"generallibrary", "generalfile", "generalvector", "generalpackager"}.difference(github.get_owners_packages()))

    def test_get_version(self):
        self.assertEqual(True, PyPI("generalpackager").get_version() > "0.2.0")

    def test_get_date(self):
        self.assertLess(PyPI("generalpackager").get_date(), Date.now())

    def test_reserve_name(self):
        pass



























from generallibrary import Ver
from generalpackager.api.local_repo import LocalRepo

import unittest


class TestLocalRepo(unittest.TestCase):
    """ Skipped tests:
        pip_install
        pip_uninstall
        unittest
        create_sdist
        upload
        get_path_from_name
        get_repo_path_parent
        write_metadata
    """

    def test_has_metadata(self):
        self.assertEqual(True, LocalRepo().has_metadata())
        self.assertEqual(False, LocalRepo("doesntexist").has_metadata())

    def test_load_metadata(self):
        self.assertEqual(True, LocalRepo().enabled)
        self.assertEqual("generalpackager", LocalRepo().name)
        self.assertIsInstance(LocalRepo().version, Ver)
        self.assertIsInstance(LocalRepo().description, str)
        self.assertIsInstance(LocalRepo().install_requires, list)
        self.assertIsInstance(LocalRepo().extras_require, dict)
        self.assertIsInstance(LocalRepo().topics, list)
        self.assertIsInstance(LocalRepo().manifest, list)

    def test_get_metadata_dict(self):
        self.assertEqual(True, LocalRepo().get_metadata_dict()["enabled"])

    def test_exists(self):
        self.assertEqual(True, LocalRepo().exists())
        self.assertEqual(True, LocalRepo.path_exists(LocalRepo().path))

    def test_get_local_repos(self):
        self.assertEqual(LocalRepo().path.get_parent(), LocalRepo.get_repos_path())

    def test_paths(self):
        method_names = (
            "get_readme_path",
            "get_metadata_path",
            "get_git_exclude_path",
            "get_setup_path",
            "get_manifest_path",
            "get_license_path",
            "get_workflow_path",
            "get_test_path",
            "get_init_path",
        )
        local_repo = LocalRepo()
        for name in method_names:
            self.assertEqual(True, getattr(local_repo, name)().exists())

    def test_get_test_paths(self):
        self.assertLess(2, len(list(LocalRepo().get_test_paths())))

    def test_text_in_tests(self):
        self.assertEqual(True, LocalRepo().text_in_tests("stringthatexists"))
        self.assertEqual(False, LocalRepo().text_in_tests("stringthat" + "doesntexist"))

    def test_get_package_paths(self):
        package_paths = list(LocalRepo().get_package_paths_gen())
        self.assertIn(LocalRepo().get_test_path(), package_paths)
        self.assertIn(LocalRepo().path / LocalRepo().name, package_paths)
        self.assertNotIn(LocalRepo().path, package_paths)

    def test_get_changed_files(self):
        local_repo = LocalRepo()
        version = local_repo.version
        local_repo.bump_version()
        self.assertNotEqual(local_repo.version, version)
        self.assertIn("metadata.json", local_repo.git_changed_files())
        local_repo.version = version
        self.assertEqual(local_repo.version, version)

    def test_get_repo_path_child(self):
        self.assertNotEqual(None, LocalRepo.get_repo_path_child(LocalRepo().path.get_parent()))
        self.assertEqual(None, LocalRepo.get_repo_path_child(LocalRepo().path))



































from generallibrary import CodeLine, comma_and_and, EnvVar

from itertools import chain


class _PackagerWorkflow:
    """ Light handling of workflow logic. """
    @staticmethod
    def _var(string):
        return f"${{{{ {string} }}}}"

    @staticmethod
    def _commit_msg_if(**conditions):
        checks = [f"contains(github.event.head_commit.message, '[CI {key}]') == {str(value).lower()}" for key, value in conditions.items()]
        return f"if: {' && '.join(checks)}"

    _commit_msg = "github.event.head_commit.message"
    _action_checkout = "actions/checkout@v2"
    _action_setup_python = "actions/setup-python@v2"
    _action_setup_ssh = "webfactory/ssh-agent@v0.5.3"
    _matrix_os = "matrix.os"
    _matrix_python_version = "matrix.python-version"

    def get_triggers(self):
        """ :param generalpackager.Packager self: """
        on = CodeLine("on:")
        push = on.add_node("push:")
        branches = push.add_node("branches:")
        branches.add_node("- master")
        return on

    def _get_step(self, name, *codelines):
        step = CodeLine(f"- name: {name}")
        for codeline in codelines:
            if codeline:
                step.add_node(codeline)
        return step

    def step_setup_ssh(self):
        """ :param generalpackager.Packager self: """
        with_ = CodeLine("with:")
        with_.add_node("ssh-private-key: ${{ secrets.GIT_SSH }}")
        return self._get_step(f"Set up Git SSH", f"uses: {self._action_setup_ssh}", with_)

    def step_setup_python(self, version):
        """ :param generalpackager.Packager self:
            :param version: """
        with_ = CodeLine("with:")
        with_.add_node(f"python-version: {version}")
        return self._get_step(f"Set up python version {version}", f"uses: {self._action_setup_python}", with_)

    def step_install_necessities(self):
        """ :param generalpackager.Packager self: """
        run = CodeLine("run: |")
        run.add_node("python -m pip install --upgrade pip")
        run.add_node(f"pip install setuptools wheel twine")
        return self._get_step(f"Install necessities pip, setuptools, wheel, twine", run)

    def step_install_package_pip(self, *packagers):
        """ Supply Packagers to create pip install steps for.

            :param generalpackager.Packager self: """
        names = [p.name for p in packagers]
        run = CodeLine(f"run: pip install {' '.join(names)}")
        return self._get_step(f"Install pip packages {comma_and_and(*names, period=False)}", run)

    def step_install_package_git(self, *packagers):
        """ Supply Packagers to create git install steps for.

            :param generalpackager.Packager self: """
        run = CodeLine(f"run: |")
        for packager in packagers:
            run.add_node(f"pip install git+ssh://git@github.com/{packager.github.owner}/{packager.name}.git")

        for packager in chain(packagers, self.summary_packagers):  # Also clone summary_packagers for syncing
            run.add_node(f"git clone ssh://git@github.com/{packager.github.owner}/{packager.name}.git")

        return self._get_step(f"Install and clone {len(packagers)} git repos", run)

    def get_env(self):
        """ :param generalpackager.Packager self: """
        env = CodeLine("env:")
        for packager in self.get_all():
            for env_var in packager.localmodule.get_env_vars():
                if env_var.actions_name and env_var.name not in str(env):
                    env.add_node(f"{env_var.name}: {env_var.actions_name}")
        if not env.get_children():
            return None
        return env

    def steps_setup(self, python_version):
        """ :param generalpackager.Packager self:
            :param python_version: """
        steps = CodeLine("steps:")
        steps.add_node(self.step_setup_ssh())
        steps.add_node(self.step_setup_python(version=python_version))
        steps.add_node(self.step_install_necessities())
        steps.add_node(self.step_install_package_git(*self.get_ordered_packagers()))
        return steps

    def get_unittest_job(self):
        """ :param generalpackager.Packager self: """
        job = CodeLine("unittest:")
        job.add_node(self._commit_msg_if(SKIP=False, AUTO=False))
        job.add_node(f"runs-on: {self._var(self._matrix_os)}")
        strategy = job.add_node("strategy:")
        matrix = strategy.add_node("matrix:")
        matrix.add_node(f"python-version: {list(self.python)}".replace("'", ""))
        matrix.add_node(f"os: {[f'{os}-latest' for os in self.os]}".replace("'", ""))

        steps = job.add_node(self.steps_setup(python_version=self._var(self._matrix_python_version)))
        steps.add_node(self.step_run_packager_method("workflow_unittest"))
        return job

    def get_sync_job(self):
        """ :param generalpackager.Packager self: """
        job = CodeLine("sync:")
        job.add_node("needs: unittest")
        job.add_node(f"runs-on: ubuntu-latest")
        steps = job.add_node(self.steps_setup(python_version=self.python[0]))
        steps.add_node(self.step_run_packager_method("workflow_sync"))
        return job

    def step_run_packager_method(self, method):
        """ :param generalpackager.Packager self:
            :param method: """
        run = CodeLine(f'run: |')
        run.add_node(f'python -c "from generalpackager import Packager; Packager(\'generalpackager\').{method}()"')
        return self._get_step(f"Run Packager method '{method}'", run, self.get_env())

    def run_ordered_methods(self, *funcs):
        """ :param generalpackager.Packager self: """
        order = self.get_ordered_packagers()
        for func in funcs:
            for packager in order:
                func(packager)

    def workflow_unittest(self):
        """ :param generalpackager.Packager self: """
        self.run_ordered_methods(
            lambda packager: packager.generate_localfiles(aesthetic=False),
            lambda packager: packager.localrepo.pip_install(),
            lambda packager: packager.localrepo.unittest(),
        )

    def workflow_sync(self):
        """ Runs in workflow once Packagers have created each LocalRepo from latest master commit.

            :param generalpackager.Packager self: """
        trigger_repo = str(EnvVar('GITHUB_REPOSITORY')).split('/')[1]
        msg1 = f"[CI AUTO] Sync triggered by {trigger_repo}"
        msg2 = f"[CI AUTO] Publish triggered by {trigger_repo}"

        self.run_ordered_methods(
            lambda packager: packager.if_publish_bump(),
            lambda packager: packager.generate_localfiles(aesthetic=True),
            lambda packager: packager.localrepo.pip_install(),
            lambda packager: packager.localrepo.unittest(),  # For good measure
            lambda packager, msg=msg1: packager.commit_and_push(message=msg, tag=False),
            lambda packager, msg=msg2: packager.if_publish_publish(message=msg),
            lambda packager: packager.sync_github_metadata(),
        )

        for packager in self.summary_packagers:
            packager.upload_package_summary(msg=msg1)

    def upload_package_summary(self, msg):
        """ :param generalpackager.Packager self: """
        self.file_secret_readme.generate()
        self.commit_and_push(message=msg)

    def if_publish_bump(self):
        """ Bump if updated and any other Packager is bumped.

            :param generalpackager.Packager self: """
        if self.general_bumped_set() and not self.is_bumped() and self.compare_local_to_pypi(aesthetic=False):
            self.localrepo.bump_version()

    def if_publish_publish(self, message):
        """ :param generalpackager.Packager self:
            :param message: """
        if self.is_bumped():
            self.file_readme.generate()
            self.commit_and_push(message=message, tag=True)
            if not self.localrepo.private:
                self.localrepo.upload()











from generallibrary import Date
from generalfile import Path

from requests import ConnectionError


class _PackagerPypi:
    def get_latest_release(self):
        """ Use current datetime if bumped, otherwise fetch.

            :param generalpackager.Packager self: """
        try:
            bumped = self.is_bumped()
        except ConnectionError:
            return "Failed fetching"

        if bumped:
            return Date.now()
        else:
            return self.pypi.get_date()

    def reserve_name(self):
        """ Reserve a name on PyPI with template files.
            Untested.

            :param generalpackager.Packager self: """
        path = Path.get_cache_dir() / "python/pypi_reserve"  # type: Path
        packager = type(self)(self.name, path=path)
        packager.create_blank_locally(install=False)
        packager.localrepo.upload()
        path.delete()
































from generalpackager.api.shared import _SharedAPI
from generallibrary import ObjInfo, deco_cache, Recycle, EnvVar, get, import_module

import pkg_resources



class LocalModule(Recycle, _SharedAPI):
    """ Tools to interface a Local Python Module. """
    _recycle_keys = {"name": lambda name: LocalModule._scrub_name(name=name)}

    def __init__(self, name=None):
        self.name = self._scrub_name(name=name)

    @classmethod
    def _scrub_name(cls, name):
        return name or "generalpackager"

    def __repr__(self):
        return self.name

    def exists(self):
        """ Return whether this API's target exists. """
        return bool(import_module(name=self.name, error=False))

    @property
    @deco_cache()
    def module(self):
        return import_module(self.name, error=False)

    def _filter(self, objInfo):
        """ :param ObjInfo objInfo: """
        return objInfo.module().__name__.startswith(self.name) and (objInfo.from_class() or objInfo.from_module()) and not objInfo.is_instance()

    @property
    @deco_cache()
    def objInfo(self):
        if self.module is None:
            return None

        objInfo = ObjInfo(self.module)
        assert objInfo.is_module()

        # objInfo.children_states = ObjInfo.children_states.copy()
        # objInfo.children_states[ObjInfo.is_instance] = False

        objInfo.get_children(depth=-1, filt=self._filter, traverse_excluded=False)
        objInfo.disconnect(lambda node: not self._filter(node))

        return objInfo

    @deco_cache()
    def get_env_vars(self):
        """ Get a list of EnvVar instances available directly in module.

            :rtype: list[generallibrary.EnvVar] """
        new_objInfo = ObjInfo(self.module)
        new_objInfo.all_identifiers = []  # Bad fix for bad circularity prevention
        return [objInfo.obj for objInfo in new_objInfo.get_children() if isinstance(objInfo.obj, EnvVar)]

    @staticmethod
    @deco_cache()
    def get_all_local_modules():
        """ Get a list of all available LocalModules. """
        return [LocalModule(name=pkg.project_name) for pkg in pkg_resources.working_set]

    @deco_cache()
    def get_dependencies(self):
        """ Get a list of LocalModules that this module depends on. """
        pkg = get(pkg_resources.working_set.by_key, self.name.lower())
        if not pkg:
            return []
        try:
            requires = pkg.requires()
        except FileNotFoundError:
            return []
        return [LocalModule(name=str(name)) for name in requires]

    @deco_cache()
    def get_dependants(self):
        """ Get a list of LocalModules that depend on this module. """
        return [local_module for local_module in self.get_all_local_modules() if self in local_module.get_dependencies()]































from generalpackager import Packager
from generalfile.test.setup_workdir import setup_workdir

import unittest


class TestPackager(unittest.TestCase):
    """ Skipped:
        spawn_children
        spawn_parents
        """
    def test_exists(self):
        self.assertEqual(True, Packager().exists())




